#!/usr/local/bin/python
# -*- coding: utf-8 -*-
# moments_models_2pop_bb_parallel_folded_2periods

######################################################################################################################################################
#								IMPORT PACKAGES
######################################################################################################################################################
import sys
import moments
from moments import Misc,Spectrum,Numerics,Manips,Integration,Demographics1D,Demographics2D
import numpy
from scipy.stats import betabinom
from multiprocess import Pool
import copy
nbcores = 10


######################################################################################################################################################
#							BETA-BINOMIAL PROBABILITY MASS FUNCTION 
######################################################################################################################################################
#https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.betabinom.html
#require Python >= 3.5, NumPy >=1.13.3, SciPy >= 1.4.0
myPlot = False

if myPlot==True:
    import matplotlib.pyplot as plt

    n, a, b = 10, 2, 2
    bb = betabinom(n, a, b)
    bbarray = [bb.pmf(i) for i in range(0,n+1)]

    sum(bbarray)
    bbarray

    fig, ax = plt.subplots(1, 1)
    ax.plot(numpy.arange(0, 1.1, 0.1), bbarray, 'go', ms=8, label='betabinom pmf')
    ax.vlines(numpy.arange(0, 1.1, 0.1), 0, bbarray, colors='g', lw=5, alpha=0.5)
    plt.show()


######################################################################################################################################################
#							STRICT ISOLATION MODELS - 2 POPULATIONS
######################################################################################################################################################
def SI(params, ns):
    nu1, nu2, Ts = params
    n1, n2 = ns
    """
    Strict Isolation.

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    Ts: Duration of divergence in isolation.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Equilibrium ancestral population
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    # Split event
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)
    # Integrate by setting divergence time to Ts, the population sizes to nu1 and nu2, and the migration rates to zero
    phi.integrate([nu1, nu2], Ts)
    # Calculate the folded spectrum
    fs = phi
    return fs


def SIexp(params, ns):
    nu1e, nu1, nu2e, nu2, Ts, Ti = params
    n1, n2 = ns
    """
    Strict Isolation with exponential size changes.

    nu1e: Size of population 1 after split.
    nu1: Size of population 1 after exponential size change.
    nu2e: Size of population 2 after split.
    nu2: Size of population 2 after exponential size change.
    Ts: Duration of divergence in isolation.
    Ti: Duration of divergence in isolation with pop1: exponential change / pop2: exponential change.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)
    phi.integrate([nu1e, nu2e], Ts)
    nu1_func = lambda t: nu1e*(nu1/nu1e) ** (t/Ti)
    nu2_func = lambda t: nu2e*(nu2/nu2e) ** (t/Ti)
    nu_func = lambda t: [nu1_func(t), nu2_func(t)]
    phi.integrate(nu_func, Ti)
    fs = phi
    return fs


def SI2N(params, ns):
    nu1, nu2, Ts, nr, bf = params
    n1, n2 = ns
    """
    Strict Isolation with heterogeneous Ne (shared between populations).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    Ts: Duration of divergence in isolation.
    nr: Fraction of non-recombining regions, i.e. with reduced Ne.
    bf : Background factor that defines to which extent Ne is reduced.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for specific demographic steps
    def calc_bin(l_bf, l_nr):
        #l_bf is either = 1 or bf
        #l_nr is either = 1-nr or nr
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1*l_bf, nu2*l_bf], Ts)
        fsBin = phiBin
        return( l_nr * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of bf and nr 
    with Pool(processes=2) as pool:
        results = pool.starmap(calc_bin, zip([bf,1], [nr,1-nr]))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs


def SIbbN(params, ns):
    nu1, nu2, Ts, a, b = params
    n1, n2 = ns
    """
    Strict Isolation with heterogeneous beta-binomial Ne (shared between populations).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    Ts: Duration of divergence in isolation.
    a, b: parameter of the Beta Binomial PMF.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Draw the fractions of SFS in each of the 10 Background Selection Factor bins from a Beta-Binomial PMF for all populations
    bb = betabinom(9, a, b)
    bbarray = [bb.pmf(0),bb.pmf(1),bb.pmf(2),bb.pmf(3),bb.pmf(4),bb.pmf(5),bb.pmf(6),bb.pmf(7),bb.pmf(8),bb.pmf(9)]
    bins = list(range(1,11))
    BSF = [x/10 for x in bins]

    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for bin-specific demographic steps
    def calc_bin(BSFb, bbVal):
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1*BSFb, nu2*BSFb], Ts)
        fsBin = phiBin
        return(bbVal * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of bin and bb
    with Pool(processes=nbcores) as pool:
        results = pool.starmap(calc_bin, zip(BSF, bbarray))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs


######################################################################################################################################################
#							ISOLATION WITH MIGRATION MODELS - 2 POPULATIONS
######################################################################################################################################################
def IM(params, ns):
    nu1, nu2, m12, m21, Ts = params
    n1, n2 = ns
    """
    Isolation with Migration.

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1.
    m21: Migration from pop 1 to pop 2.
    Ts: Duration of divergence with migration.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)
    phi.integrate([nu1, nu2], Ts, m=numpy.array([[0, m12],[m21, 0]]))
    fs = phi
    return fs


def IMexp(params, ns):
    nu1e, nu1, nu2e, nu2, m12, m21, Ts, Ti = params
    n1, n2 = ns
    """
    Isolation with Migration with exponential size changes.

    nu1e: Size of population 1 after split.
    nu1: Size of population 1 after exponential size change.
    nu2e: Size of population 2 after split.
    nu2: Size of population 2 after exponential size change.
    m12: Migration from pop 2 to pop 1.
    m21: Migration from pop 1 to pop 2.
    Ts: Duration of divergence with migration.
    Ti: Duration of divergence with migration with pop1: exponential change / pop2: exponential change.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)
    phi.integrate([nu1e, nu2e], Ts, m=numpy.array([[0, m12],[m21, 0]]))
    nu1_func = lambda t: nu1e*(nu1/nu1e) ** (t/Ti)
    nu2_func = lambda t: nu2e*(nu2/nu2e) ** (t/Ti)
    nu_func = lambda t: [nu1_func(t), nu2_func(t)]
    phi.integrate(nu_func, Ti, m=numpy.array([[0, m12],[m21, 0]]))
    fs = phi
    return fs


def IM2N(params, ns):
    nu1, nu2, m12, m21, Ts, nr, bf = params
    n1, n2 = ns
    """
    Isolation with Migration with heterogeneous Ne (shared between populations).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1.
    m21: Migration from pop 1 to pop 2.
    Ts: Duration of divergence with migration.
    nr: Fraction of non-recombining regions, i.e. with reduced Ne.
    bf : Background factor that defines to which extent Ne is reduced.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for specific demographic steps
    def calc_bin(l_bf, l_nr):
        #l_bf is either = 1 or bf
        #l_nr is either = 1-nr or nr
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1*l_bf, nu2*l_bf], Ts, m=numpy.array([[0, m12],[m21, 0]]))
        fsBin = phiBin
        return( l_nr * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of bf and nr 
    with Pool(processes=2) as pool:
        results = pool.starmap(calc_bin, zip([bf,1], [nr,1-nr]))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs 


def IMbbN(params, ns):
    nu1, nu2, m12, m21, Ts, a, b = params
    n1, n2 = ns
    """
    Isolation with Migration with heterogeneous beta-binomial Ne (shared between populations).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1.
    m21: Migration from pop 1 to pop 2.
    Ts: Duration of divergence with migration.
    a, b: parameter of the Beta Binomial PMF.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Draw the fractions of SFS in each of the 10 Background Selection Factor bins from a Beta-Binomial PMF for all populations
    bb = betabinom(9, a, b)
    bbarray = [bb.pmf(0),bb.pmf(1),bb.pmf(2),bb.pmf(3),bb.pmf(4),bb.pmf(5),bb.pmf(6),bb.pmf(7),bb.pmf(8),bb.pmf(9)]
    bins = list(range(1,11))
    BSF = [x/10 for x in bins]

    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for bin-specific demographic steps
    def calc_bin(BSFb, bbVal):
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1*BSFb, nu2*BSFb], Ts, m=numpy.array([[0, m12],[m21, 0]]))
        fsBin = phiBin
        return(bbVal * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of bin and bb 
    with Pool(processes=nbcores) as pool:
        results = pool.starmap(calc_bin, zip(BSF, bbarray))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs


def IM2M(params, ns):
    nu1, nu2, m12, m21, Ts, P = params
    n1, n2 = ns
    """
    Isolation with Migration with heterogeneous Me (shared between populations).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1 in neutral regions. In barrier regions, m12=0.
    m21: Migration from pop 1 to pop 2 in neutral regions. In barrier regions, m21=0.
    Ts: Duration of divergence with migration.
    P: Fraction of the genome evolving neutrally.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for specific demographic steps
    def calc_bin(l_i, l_P):
        #l_i is either = 1 or 0
        #l_P is either = P or 1-P
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1, nu2], Ts, m=numpy.array([[0, m12*l_i],[m21*l_i, 0]]))
        fsBin = phiBin
        return( l_P * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of i and P
    with Pool(processes=2) as pool:
        results = pool.starmap(calc_bin, zip([1,0], [P,1-P]))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs 
  

def IMbbM(params, ns):
    nu1, nu2, m12, m21, Ts, a, b = params
    n1, n2 = ns
    """
    Isolation with Migration with heterogeneous beta-binomial Me (shared between populations).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1.
    m21: Migration from pop 1 to pop 2.
    Ts: Duration of divergence with migration.
    a, b: parameter of the Beta Binomial PMF.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Draw the fractions of SFS in each of the 11 Gene-Flow Factor bins from a Beta-Binomial PMF for all populations
    bb = betabinom(10, a, b)
    bbarray = [bb.pmf(0),bb.pmf(1),bb.pmf(2),bb.pmf(3),bb.pmf(4),bb.pmf(5),bb.pmf(6),bb.pmf(7),bb.pmf(8),bb.pmf(9),bb.pmf(10)]
    bins = list(range(0,11))
    GFF = [x/10 for x in bins]

    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for bin-specific demographic steps
    def calc_bin(GFFb, bbVal):
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1, nu2], Ts, m=numpy.array([[0, m12*GFFb],[m21*GFFb, 0]]))
        fsBin = phiBin
        return(bbVal * fsBin)

    # Run the calc_bin function with a pool of nbcores+1 agents - each core will have a value of bin and bb 
    with Pool(processes=nbcores+1) as pool:
        results = pool.starmap(calc_bin, zip(GFF, bbarray))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs


def IM2M2P(params, ns):
    nu1, nu2, m12, m21, Ts, P1, P2 = params
    n1, n2 = ns
    """
    Isolation with Migration with heterogeneous Me (population-specific).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1 in neutral regions. In barrier regions, m12=0.
    m21: Migration from pop 1 to pop 2 in neutral regions. In barrier regions, m21=0.
    Ts: Duration of divergence with migration.
    P1: Fraction of the genome evolving neutrally in population 1.
    P2: Fraction of the genome evolving neutrally in population 2.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for specific demographic steps
    def calc_bin(l_i1, l_i2, l_P1, l_P2):
        #l_i1 is either = 1 or 0
        #l_i2 is either = 1 or 0
        #l_P1 is either = P1 or 1-P1
        #l_P2 is either = P2 or 1-P2
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1, nu2], Ts, m=numpy.array([[0, m12*l_i1],[m21*l_i2, 0]]))
        fsBin = phiBin
        return( l_P1 * l_P2 * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of i1, i2, P1 and P2
    with Pool(processes=4) as pool:
        results = pool.starmap(calc_bin, zip([1,0,1,0],[1,0,0,1],[P1,1-P1,P1,1-P1],[P2,1-P2,1-P2,P2]))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs


def IM2N2M(params, ns):
    nu1, nu2, m12, m21, Ts, nr, bf, P = params
    n1, n2 = ns
    """
    Isolation with Migration with heterogeneous Ne (shared between populations) and heterogeneous Me (shared between populations).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1 in neutral regions. In barrier regions, m12=0.
    m21: Migration from pop 1 to pop 2 in neutral regions. In barrier regions, m21=0.
    Ts: Duration of divergence with migration.
    nr: Fraction of non-recombining regions, i.e. with reduced Ne.
    bf : Background factor that defines to which extent Ne is reduced.
    P: Fraction of the genome evolving neutrally.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for specific demographic steps
    def calc_bin(l_bf, l_nr, l_i, l_P):
        #l_bf is either = 1 or bf
        #l_nr is either = 1-nr or nr
        #l_i is either = 1 or 0
        #l_P is either = P or 1-P
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1*l_bf, nu2*l_bf], Ts, m=numpy.array([[0, m12*l_i],[m21*l_i, 0]]))
        fsBin = phiBin
        return( l_nr * l_P * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of bf, nr, i and P
    with Pool(processes=4) as pool:
        results = pool.starmap(calc_bin, zip([bf,1,bf,1], [nr,1-nr,nr,1-nr], [1,1,0,0], [P,P,1-P,1-P]))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs 


def IM2N2M2P(params, ns):
    nu1, nu2, m12, m21, Ts, nr, bf, P1, P2 = params
    n1, n2 = ns
    """
    Isolation with Migration with heterogeneous Ne (shared between populations) and heterogeneous Me (population-specific).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1 in neutral regions. In barrier regions, m12=0.
    m21: Migration from pop 1 to pop 2 in neutral regions. In barrier regions, m21=0.
    Ts: Duration of divergence with migration.
    nr: Fraction of non-recombining regions, i.e. with reduced Ne.
    bf : Background factor that defines to which extent Ne is reduced.
    P1: Fraction of the genome evolving neutrally in population 1.
    P2: Fraction of the genome evolving neutrally in population 2.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for specific demographic steps
    def calc_bin(l_bf, l_nr, l_i1, l_i2, l_P1, l_P2):
        #l_bf is either = 1 or bf
        #l_nr is either = 1-nr or nr
        #l_i1 is either = 1 or 0
        #l_i2 is either = 1 or 0
        #l_P1 is either = P1 or 1-P1
        #l_P2 is either = P2 or 1-P2
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1*l_bf, nu2*l_bf], Ts, m=numpy.array([[0, m12*l_i1],[m21*l_i2, 0]]))
        fsBin = phiBin
        return( l_nr * l_P1 * l_P2 * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of bf, nr, i1, i2, P1 and P2
    with Pool(processes=8) as pool:
        results = pool.starmap(calc_bin, zip([bf,1,bf,1,bf,1,bf,1], [nr,1-nr,nr,1-nr,nr,1-nr,nr,1-nr], [1,1,0,0,1,1,0,0], [1,1,0,0,0,0,1,1], [P1,P1,1-P1,1-P1,P1,P1,1-P1,1-P1], [P2,P2,1-P2,1-P2,1-P2,1-P2,P2,P2]))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs


######################################################################################################################################################
#							SECONDARY CONTACT MODELS - 2 POPULATIONS
######################################################################################################################################################
def SC(params, ns):
    nu1, nu2, m12, m21, Ts, Ti = params
    n1, n2 = ns
    """
    Secondary Contact.

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1.
    m21: Migration from pop 1 to pop 2.
    Ts: Duration of divergence in isolation.
    Ti: Duration of divergence with migration.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)
    phi.integrate([nu1, nu2], Ts, m=numpy.array([[0, 0],[0, 0]]))
    phi.integrate([nu1, nu2], Ti, m=numpy.array([[0, m12],[m21, 0]]))
    fs = phi
    return fs


def SCexp(params, ns):
    nu1e, nu1, nu2e, nu2, m12, m21, Ts, Ti = params
    n1, n2 = ns
    """
    Secondary Contact with exponential size changes.

    nu1e: Size of population 1 after split.
    nu1: Size of population 1 after exponential size change.
    nu2e: Size of population 2 after split.
    nu2: Size of population 2 after exponential size change.
    m12: Migration from pop 2 to pop 1.
    m21: Migration from pop 1 to pop 2.
    Ts: Duration of divergence in isolation.
    Ti: Duration of divergence with migration with pop1: exponential change / pop2: exponential change.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)
    phi.integrate([nu1e, nu2e], Ts, m=numpy.array([[0, 0],[0, 0]]))
    nu1_func = lambda t: nu1e*(nu1/nu1e) ** (t/Ti)
    nu2_func = lambda t: nu2e*(nu2/nu2e) ** (t/Ti)
    nu_func = lambda t: [nu1_func(t), nu2_func(t)]
    phi.integrate(nu_func, Ti, m=numpy.array([[0, m12],[m21, 0]]))
    fs = phi
    return fs


def SC2N(params, ns):
    nu1, nu2, m12, m21, Ts, Ti, nr, bf = params
    n1, n2 = ns
    """
    Secondary Contact with heterogeneous Ne (shared between populations).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1.
    m21: Migration from pop 1 to pop 2.
    Ts: Duration of divergence in isolation.
    Ti: Duration of divergence with migration.
    nr: Fraction of non-recombining regions, i.e. with reduced Ne.
    bf : Background factor that defines to which extent Ne is reduced.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for specific demographic steps
    def calc_bin(l_bf, l_nr):
        #l_bf is either = 1 or bf
        #l_nr is either = 1-nr or nr
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1*l_bf, nu2*l_bf], Ts, m=numpy.array([[0, 0],[0, 0]]))
        phiBin.integrate([nu1*l_bf, nu2*l_bf], Ti, m=numpy.array([[0, m12],[m21, 0]]))
        fsBin = phiBin
        return( l_nr * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of bf and nr 
    with Pool(processes=2) as pool:
        results = pool.starmap(calc_bin, zip([bf,1], [nr,1-nr]))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs


def SCbbN(params, ns):
    nu1, nu2, m12, m21, Ts, Ti, a, b = params
    n1, n2 = ns
    """
    Secondary Contact with heterogeneous beta-binomial Ne (shared between populations).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1.
    m21: Migration from pop 1 to pop 2.
    Ts: Duration of divergence in isolation.
    Ti: Duration of divergence with migration.
    a, b: parameter of the Beta Binomial PMF.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Draw the fractions of SFS in each of the 10 Background Selection Factor bins from a Beta-Binomial PMF for all populations
    bb = betabinom(9, a, b)
    bbarray = [bb.pmf(0),bb.pmf(1),bb.pmf(2),bb.pmf(3),bb.pmf(4),bb.pmf(5),bb.pmf(6),bb.pmf(7),bb.pmf(8),bb.pmf(9)]
    bins = list(range(1,11))
    BSF = [x/10 for x in bins]

    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for bin-specific demographic steps
    def calc_bin(BSFb, bbVal):
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1*BSFb, nu2*BSFb], Ts, m=numpy.array([[0, 0],[0, 0]]))
        phiBin.integrate([nu1*BSFb, nu2*BSFb], Ti, m=numpy.array([[0, m12],[m21, 0]]))
        fsBin = phiBin
        return(bbVal * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of bin and bb 
    with Pool(processes=nbcores) as pool:
        results = pool.starmap(calc_bin, zip(BSF, bbarray))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs


def SC2M(params, ns):
    nu1, nu2, m12, m21, Ts, Ti, P = params
    n1, n2 = ns
    """
    Secondary Contact with heterogeneous Me (shared between populations).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1 in neutral regions. In barrier regions, m12=0.
    m21: Migration from pop 1 to pop 2 in neutral regions. In barrier regions, m21=0.
    Ts: Duration of divergence in isolation.
    Ti: Duration of divergence with migration.
    P: Fraction of the genome evolving neutrally.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for specific demographic steps
    def calc_bin(l_i, l_P):
        #l_i is either = 1 or 0
        #l_P is either = P or 1-P
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1, nu2], Ts, m=numpy.array([[0, 0],[0, 0]]))
        phiBin.integrate([nu1, nu2], Ti, m=numpy.array([[0, m12*l_i],[m21*l_i, 0]]))
        fsBin = phiBin
        return( l_P * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of i and P
    with Pool(processes=2) as pool:
        results = pool.starmap(calc_bin, zip([1,0], [P,1-P]))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs 


def SCbbM(params, ns):
    nu1, nu2, m12, m21, Ts, Ti, a, b = params
    n1, n2 = ns
    """
    Secondary Contact with heterogeneous beta-binomial Me (shared between populations).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1.
    m21: Migration from pop 1 to pop 2.
    Ts: Duration of divergence in isolation.
    Ti: Duration of divergence with migration.
    a, b: parameter of the Beta Binomial PMF.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Draw the fractions of SFS in each of the 11 Gene-Flow Factor bins from a Beta-Binomial PMF for all populations
    bb = betabinom(10, a, b)
    bbarray = [bb.pmf(0),bb.pmf(1),bb.pmf(2),bb.pmf(3),bb.pmf(4),bb.pmf(5),bb.pmf(6),bb.pmf(7),bb.pmf(8),bb.pmf(9),bb.pmf(10)]
    bins = list(range(0,11))
    GFF = [x/10 for x in bins]

    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for bin-specific demographic steps
    def calc_bin(GFFb, bbVal):
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1, nu2], Ts, m=numpy.array([[0, 0],[0, 0]]))
        phiBin.integrate([nu1, nu2], Ti, m=numpy.array([[0, m12*GFFb],[m21*GFFb, 0]]))
        fsBin = phiBin
        return(bbVal * fsBin)

    # Run the calc_bin function with a pool of nbcores+1 agents - each core will have a value of bin and bb 
    with Pool(processes=nbcores+1) as pool:
        results = pool.starmap(calc_bin, zip(GFF, bbarray))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs


def SC2M2P(params, ns):
    nu1, nu2, m12, m21, Ts, Ti, P1, P2 = params
    n1, n2 = ns
    """
    Secondary Contact with heterogeneous Me (population-specific).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1 in neutral regions. In barrier regions, m12=0.
    m21: Migration from pop 1 to pop 2 in neutral regions. In barrier regions, m21=0.
    Ts: Duration of divergence in isolation.
    Ti: Duration of divergence with migration.
    P1: Fraction of the genome evolving neutrally in population 1.
    P2: Fraction of the genome evolving neutrally in population 2.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for specific demographic steps
    def calc_bin(l_i1, l_i2, l_P1, l_P2):
        #l_i1 is either = 1 or 0
        #l_i2 is either = 1 or 0
        #l_P1 is either = P1 or 1-P1
        #l_P2 is either = P2 or 1-P2
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1, nu2], Ts, m=numpy.array([[0, 0],[0, 0]]))
        phiBin.integrate([nu1, nu2], Ti, m=numpy.array([[0, m12*l_i1],[m21*l_i2, 0]]))
        fsBin = phiBin
        return( l_P1 * l_P2 * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of i1, i2, P1 and P2
    with Pool(processes=4) as pool:
        results = pool.starmap(calc_bin, zip([1,0,1,0],[1,0,0,1],[P1,1-P1,P1,1-P1],[P2,1-P2,1-P2,P2]))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs


def SC2N2M(params, ns):
    nu1, nu2, m12, m21, Ts, Ti, nr, bf, P = params
    n1, n2 = ns
    """
    Secondary Contact with heterogeneous Ne (shared between populations) and heterogeneous Me (shared between populations).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1 in neutral regions. In barrier regions, m12=0.
    m21: Migration from pop 1 to pop 2 in neutral regions. In barrier regions, m21=0.
    Ts: Duration of divergence in isolation.
    Ti: Duration of divergence with migration.
    nr: Fraction of non-recombining regions, i.e. with reduced Ne.
    bf : Background factor that defines to which extent Ne is reduced.
    P: Fraction of the genome evolving neutrally.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for specific demographic steps
    def calc_bin(l_bf, l_nr, l_i, l_P):
        #l_bf is either = 1 or bf
        #l_nr is either = 1-nr or nr
        #l_i is either = 1 or 0
        #l_P is either = P or 1-P
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1*l_bf, nu2*l_bf], Ts, m=numpy.array([[0, 0],[0, 0]]))
        phiBin.integrate([nu1*l_bf, nu2*l_bf], Ti, m=numpy.array([[0, m12*l_i],[m21*l_i, 0]]))
        fsBin = phiBin
        return( l_nr * l_P * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of bf, nr, i and P
    with Pool(processes=4) as pool:
        results = pool.starmap(calc_bin, zip([bf,1,bf,1], [nr,1-nr,nr,1-nr], [1,1,0,0], [P,P,1-P,1-P]))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs 


def SC2N2M2P(params, ns):
    nu1, nu2, m12, m21, Ts, Ti, nr, bf, P1, P2 = params
    n1, n2 = ns
    """
    Secondary Contact with heterogeneous Ne (shared between populations) and heterogeneous Me (population-specific).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1 in neutral regions. In barrier regions, m12=0.
    m21: Migration from pop 1 to pop 2 in neutral regions. In barrier regions, m21=0.
    Ts: Duration of divergence in isolation.
    Ti: Duration of divergence with migration.
    nr: Fraction of non-recombining regions, i.e. with reduced Ne.
    bf : Background factor that defines to which extent Ne is reduced.
    P1: Fraction of the genome evolving neutrally in population 1.
    P2: Fraction of the genome evolving neutrally in population 2.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for specific demographic steps
    def calc_bin(l_bf, l_nr, l_i1, l_i2, l_P1, l_P2):
        #l_bf is either = 1 or bf
        #l_nr is either = 1-nr or nr
        #l_i1 is either = 1 or 0
        #l_i2 is either = 1 or 0
        #l_P1 is either = P1 or 1-P1
        #l_P2 is either = P2 or 1-P2
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1*l_bf, nu2*l_bf], Ts, m=numpy.array([[0, 0],[0, 0]]))
        phiBin.integrate([nu1*l_bf, nu2*l_bf], Ti, m=numpy.array([[0, m12*l_i1],[m21*l_i2, 0]]))
        fsBin = phiBin
        return( l_nr * l_P1 * l_P2 * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of bf, nr, i1, i2, P1 and P2
    with Pool(processes=8) as pool:
        results = pool.starmap(calc_bin, zip([bf,1,bf,1,bf,1,bf,1], [nr,1-nr,nr,1-nr,nr,1-nr,nr,1-nr], [1,1,0,0,1,1,0,0], [1,1,0,0,0,0,1,1], [P1,P1,1-P1,1-P1,P1,P1,1-P1,1-P1], [P2,P2,1-P2,1-P2,1-P2,1-P2,P2,P2]))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs


######################################################################################################################################################
#							ANCIENT MIGRATION MODELS - 2 POPULATIONS
######################################################################################################################################################
def AM(params, ns):
    nu1, nu2, m12, m21, Ts, Ti = params
    n1, n2 = ns
    """
    Ancient Migration.

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1.
    m21: Migration from pop 1 to pop 2.
    Ts: Duration of divergence with migration.
    Ti: Duration of divergence in isolation.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)
    phi.integrate([nu1, nu2], Ts, m=numpy.array([[0, m12],[m21, 0]]))
    phi.integrate([nu1, nu2], Ti, m=numpy.array([[0, 0],[0, 0]]))
    fs = phi
    return fs


def AMexp(params, ns):
    nu1e, nu1, nu2e, nu2, m12, m21, Ts, Ti = params
    n1, n2 = ns
    """
    Ancient Migration with exponential size changes.

    nu1e: Size of population 1 after split.
    nu1: Size of population 1 after exponential size change.
    nu2e: Size of population 2 after split.
    nu2: Size of population 2 after exponential size change.
    m12: Migration from pop 2 to pop 1.
    m21: Migration from pop 1 to pop 2.
    Ts: Duration of divergence with migration.
    Ti: Duration of divergence in isolation with pop1: exponential change / pop2: exponential change.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)
    phi.integrate([nu1e, nu2e], Ts, m=numpy.array([[0, m12],[m21, 0]]))
    nu1_func = lambda t: nu1e*(nu1/nu1e) ** (t/Ti)
    nu2_func = lambda t: nu2e*(nu2/nu2e) ** (t/Ti)
    nu_func = lambda t: [nu1_func(t), nu2_func(t)]
    phi.integrate(nu_func, Ti, m=numpy.array([[0, 0],[0, 0]]))
    fs = phi
    return fs


def AM2N(params, ns):
    nu1, nu2, m12, m21, Ts, Ti, nr, bf = params
    n1, n2 = ns
    """
    Ancient Migration with heterogeneous Ne (shared between populations).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1.
    m21: Migration from pop 1 to pop 2.
    Ts: Duration of divergence with migration.
    Ti: Duration of divergence in isolation.
    nr: Fraction of non-recombining regions, i.e. with reduced Ne.
    bf : Background factor that defines to which extent Ne is reduced.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for specific demographic steps
    def calc_bin(l_bf, l_nr):
        #l_bf is either = 1 or bf
        #l_nr is either = 1-nr or nr
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1*l_bf, nu2*l_bf], Ts, m=numpy.array([[0, m12],[m21, 0]]))
        phiBin.integrate([nu1*l_bf, nu2*l_bf], Ti, m=numpy.array([[0, 0],[0, 0]]))
        fsBin = phiBin
        return( l_nr * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of bf and nr 
    with Pool(processes=2) as pool:
        results = pool.starmap(calc_bin, zip([bf,1], [nr,1-nr]))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs 


def AMbbN(params, ns):
    nu1, nu2, m12, m21, Ts, Ti, a, b = params
    n1, n2 = ns
    """
    Ancient Migration with heterogeneous beta-binomial Ne (shared between populations).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1.
    m21: Migration from pop 1 to pop 2.
    Ts: Duration of divergence with migration.
    Ti: Duration of divergence in isolation.
    a, b: parameter of the Beta Binomial PMF.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Draw the fractions of SFS in each of the 10 Background Selection Factor bins from a Beta-Binomial PMF for all populations
    bb = betabinom(9, a, b)
    bbarray = [bb.pmf(0),bb.pmf(1),bb.pmf(2),bb.pmf(3),bb.pmf(4),bb.pmf(5),bb.pmf(6),bb.pmf(7),bb.pmf(8),bb.pmf(9)]
    bins = list(range(1,11))
    BSF = [x/10 for x in bins]

    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for bin-specific demographic steps
    def calc_bin(BSFb, bbVal):
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1*BSFb, nu2*BSFb], Ts, m=numpy.array([[0, m12],[m21, 0]]))
        phiBin.integrate([nu1*BSFb, nu2*BSFb], Ti, m=numpy.array([[0, 0],[0, 0]]))
        fsBin = phiBin
        return(bbVal * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of bin and bb 
    with Pool(processes=nbcores) as pool:
        results = pool.starmap(calc_bin, zip(BSF, bbarray))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs


def AM2M(params, ns):
    nu1, nu2, m12, m21, Ts, Ti, P = params
    n1, n2 = ns
    """
    Ancient Migration with heterogeneous Me (shared between populations).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1 in neutral regions. In barrier regions, m12=0.
    m21: Migration from pop 1 to pop 2 in neutral regions. In barrier regions, m21=0.
    Ts: Duration of divergence with migration.
    Ti: Duration of divergence in isolation.
    P: Fraction of the genome evolving neutrally.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for specific demographic steps
    def calc_bin(l_i, l_P):
        #l_i is either = 1 or 0
        #l_P is either = P or 1-P
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1, nu2], Ts, m=numpy.array([[0, m12*l_i],[m21*l_i, 0]]))
        phiBin.integrate([nu1, nu2], Ti, m=numpy.array([[0, 0],[0, 0]]))
        fsBin = phiBin
        return( l_P * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of i and P
    with Pool(processes=2) as pool:
        results = pool.starmap(calc_bin, zip([1,0], [P,1-P]))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs 


def AMbbM(params, ns):
    nu1, nu2, m12, m21, Ts, Ti, a, b = params
    n1, n2 = ns
    """
    Ancient Migration with heterogeneous beta-binomial Me (shared between populations).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1.
    m21: Migration from pop 1 to pop 2.
    Ts: Duration of divergence with migration.
    Ti: Duration of divergence in isolation.
    a, b: parameter of the Beta Binomial PMF.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Draw the fractions of SFS in each of the 11 Gene-Flow Factor bins from a Beta-Binomial PMF for all populations
    bb = betabinom(10, a, b)
    bbarray = [bb.pmf(0),bb.pmf(1),bb.pmf(2),bb.pmf(3),bb.pmf(4),bb.pmf(5),bb.pmf(6),bb.pmf(7),bb.pmf(8),bb.pmf(9),bb.pmf(10)]
    bins = list(range(0,11))
    GFF = [x/10 for x in bins]

    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for bin-specific demographic steps
    def calc_bin(GFFb, bbVal):
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1, nu2], Ts, m=numpy.array([[0, m12*GFFb],[m21*GFFb, 0]]))
        phiBin.integrate([nu1, nu2], Ti, m=numpy.array([[0, 0],[0, 0]]))
        fsBin = phiBin
        return(bbVal * fsBin)

    # Run the calc_bin function with a pool of nbcores+1 agents - each core will have a value of bin and bb 
    with Pool(processes=nbcores+1) as pool:
        results = pool.starmap(calc_bin, zip(GFF, bbarray))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs


def AM2M2P(params, ns):
    nu1, nu2, m12, m21, Ts, Ti, P1, P2 = params
    n1, n2 = ns
    """
    Ancient Migration with heterogeneous Me (population-specific).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1 in neutral regions. In barrier regions, m12=0.
    m21: Migration from pop 1 to pop 2 in neutral regions. In barrier regions, m21=0.
    Ts: Duration of divergence with migration.
    Ti: Duration of divergence in isolation.
    P1: Fraction of the genome evolving neutrally in population 1.
    P2: Fraction of the genome evolving neutrally in population 2.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for specific demographic steps
    def calc_bin(l_i1, l_i2, l_P1, l_P2):
        #l_i1 is either = 1 or 0
        #l_i2 is either = 1 or 0
        #l_P1 is either = P1 or 1-P1
        #l_P2 is either = P2 or 1-P2
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1, nu2], Ts, m=numpy.array([[0, m12*l_i1],[m21*l_i2, 0]]))
        phiBin.integrate([nu1, nu2], Ti, m=numpy.array([[0, 0],[0, 0]]))
        fsBin = phiBin
        return( l_P1 * l_P2 * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of i1, i2, P1 and P2
    with Pool(processes=4) as pool:
        results = pool.starmap(calc_bin, zip([1,0,1,0],[1,0,0,1],[P1,1-P1,P1,1-P1],[P2,1-P2,1-P2,P2]))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs


def AM2N2M(params, ns):
    nu1, nu2, m12, m21, Ts, Ti, nr, bf, P = params
    n1, n2 = ns
    """
    Ancient Migration with heterogeneous Ne (shared between populations) and heterogeneous Me (shared between populations).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1 in neutral regions. In barrier regions, m12=0.
    m21: Migration from pop 1 to pop 2 in neutral regions. In barrier regions, m21=0.
    Ts: Duration of divergence with migration.
    Ti: Duration of divergence in isolation.
    nr: Fraction of non-recombining regions, i.e. with reduced Ne.
    bf : Background factor that defines to which extent Ne is reduced.
    P: Fraction of the genome evolving neutrally.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for specific demographic steps
    def calc_bin(l_bf, l_nr, l_i, l_P):
        #l_bf is either = 1 or bf
        #l_nr is either = 1-nr or nr
        #l_i is either = 1 or 0
        #l_P is either = P or 1-P
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1*l_bf, nu2*l_bf], Ts, m=numpy.array([[0, m12*l_i],[m21*l_i, 0]]))
        phiBin.integrate([nu1*l_bf, nu2*l_bf], Ti, m=numpy.array([[0, 0],[0, 0]]))
        fsBin = phiBin
        return( l_nr * l_P * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of bf, nr, i and P
    with Pool(processes=4) as pool:
        results = pool.starmap(calc_bin, zip([bf,1,bf,1], [nr,1-nr,nr,1-nr], [1,1,0,0], [P,P,1-P,1-P]))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs 


def AM2N2M2P(params, ns):
    nu1, nu2, m12, m21, Ts, Ti, nr, bf, P1, P2 = params
    n1, n2 = ns
    """
    Ancient Migration with heterogeneous Ne (shared between populations) and heterogeneous Me (population-specific).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1 in neutral regions. In barrier regions, m12=0.
    m21: Migration from pop 1 to pop 2 in neutral regions. In barrier regions, m21=0.
    Ts: Duration of divergence with migration.
    Ti: Duration of divergence in isolation.
    nr: Fraction of non-recombining regions, i.e. with reduced Ne.
    bf : Background factor that defines to which extent Ne is reduced.
    P1: Fraction of the genome evolving neutrally in population 1.
    P2: Fraction of the genome evolving neutrally in population 2.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for specific demographic steps
    def calc_bin(l_bf, l_nr, l_i1, l_i2, l_P1, l_P2):
        #l_bf is either = 1 or bf
        #l_nr is either = 1-nr or nr
        #l_i1 is either = 1 or 0
        #l_i2 is either = 1 or 0
        #l_P1 is either = P1 or 1-P1
        #l_P2 is either = P2 or 1-P2
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1*l_bf, nu2*l_bf], Ts, m=numpy.array([[0, m12*l_i1],[m21*l_i2, 0]]))
        phiBin.integrate([nu1*l_bf, nu2*l_bf], Ti, m=numpy.array([[0, 0],[0, 0]]))
        fsBin = phiBin
        return( l_nr * l_P1 * l_P2 * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of bf, nr, i1, i2, P1 and P2
    with Pool(processes=8) as pool:
        results = pool.starmap(calc_bin, zip([bf,1,bf,1,bf,1,bf,1], [nr,1-nr,nr,1-nr,nr,1-nr,nr,1-nr], [1,1,0,0,1,1,0,0], [1,1,0,0,0,0,1,1], [P1,P1,1-P1,1-P1,P1,P1,1-P1,1-P1], [P2,P2,1-P2,1-P2,1-P2,1-P2,P2,P2]))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs


######################################################################################################################################################
#							PERIODIC MIGRATION MODELS - 2 POPULATIONS
######################################################################################################################################################
def PMIG(params, ns):
    nu1, nu2, m12, m21, Ts, Ti, Tj = params
    n1, n2 = ns
    """
    Three-period Migration.

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1.
    m21: Migration from pop 1 to pop 2.
    Ts: Duration of divergence with migration - first period.
    Ti: Duration of divergence in isolation - second period.
    Tj: Duration of divergence with migration - third period.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)
    phi.integrate([nu1, nu2], Ts, m=numpy.array([[0, m12],[m21, 0]]))
    phi.integrate([nu1, nu2], Ti, m=numpy.array([[0, 0],[0, 0]]))
    phi.integrate([nu1, nu2], Tj, m=numpy.array([[0, m12],[m21, 0]]))
    fs = phi
    return fs


def PMIGbbN(params, ns):
    nu1, nu2, m12, m21, Ts, Ti, Tj, a, b = params
    n1, n2 = ns
    """
    Three-period Migration with heterogeneous beta-binomial Ne (shared between populations).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1.
    m21: Migration from pop 1 to pop 2.
    Ts: Duration of divergence with migration - first period.
    Ti: Duration of divergence in isolation - second period.
    Tj: Duration of divergence with migration - third period.
    a, b: parameter of the Beta Binomial PMF.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Draw the fractions of SFS in each of the 10 Background Selection Factor bins from a Beta-Binomial PMF for all populations
    bb = betabinom(9, a, b)
    bbarray = [bb.pmf(0),bb.pmf(1),bb.pmf(2),bb.pmf(3),bb.pmf(4),bb.pmf(5),bb.pmf(6),bb.pmf(7),bb.pmf(8),bb.pmf(9)]
    bins = list(range(1,11))
    BSF = [x/10 for x in bins]

    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for bin-specific demographic steps
    def calc_bin(BSFb, bbVal):
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1*BSFb, nu2*BSFb], Ts, m=numpy.array([[0, m12],[m21, 0]]))
        phiBin.integrate([nu1*BSFb, nu2*BSFb], Ti, m=numpy.array([[0, 0],[0, 0]]))
        phiBin.integrate([nu1*BSFb, nu2*BSFb], Tj, m=numpy.array([[0, m12],[m21, 0]]))
        fsBin = phiBin
        return(bbVal * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of bin and bb 
    with Pool(processes=nbcores) as pool:
        results = pool.starmap(calc_bin, zip(BSF, bbarray))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs


def PMIGbbM(params, ns):
    nu1, nu2, m12, m21, Ts, Ti, Tj, a, b = params
    n1, n2 = ns
    """
    Three-period Migration with heterogeneous beta-binomial Me (shared between populations).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1.
    m21: Migration from pop 1 to pop 2.
    Ts: Duration of divergence with migration - first period.
    Ti: Duration of divergence in isolation - second period.
    Tj: Duration of divergence with migration - third period.
    a, b: parameter of the Beta Binomial PMF.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Draw the fractions of SFS in each of the 11 Gene-Flow Factor bins from a Beta-Binomial PMF for all populations
    bb = betabinom(10, a, b)
    bbarray = [bb.pmf(0),bb.pmf(1),bb.pmf(2),bb.pmf(3),bb.pmf(4),bb.pmf(5),bb.pmf(6),bb.pmf(7),bb.pmf(8),bb.pmf(9),bb.pmf(10)]
    bins = list(range(0,11))
    GFF = [x/10 for x in bins]

    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for bin-specific demographic steps
    def calc_bin(GFFb, bbVal):
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1, nu2], Ts, m=numpy.array([[0, m12*GFFb],[m21*GFFb, 0]]))
        phiBin.integrate([nu1, nu2], Ti, m=numpy.array([[0, 0],[0, 0]]))
        phiBin.integrate([nu1, nu2], Tj, m=numpy.array([[0, m12*GFFb],[m21*GFFb, 0]]))
        fsBin = phiBin
        return(bbVal * fsBin)

    # Run the calc_bin function with a pool of nbcores+1 agents - each core will have a value of bin and bb 
    with Pool(processes=nbcores+1) as pool:
        results = pool.starmap(calc_bin, zip(GFF, bbarray))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs


def PMIG2N2M(params, ns):
    nu1, nu2, m12, m21, Ts, Ti, Tj, nr, bf, P = params
    n1, n2 = ns
    """
    Three-period Migration with heterogeneous Ne (shared between populations) and heterogeneous Me (shared between populations).

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from pop 2 to pop 1 in neutral regions. In barrier regions, m12=0.
    m21: Migration from pop 1 to pop 2 in neutral regions. In barrier regions, m21=0.
    Ts: Duration of divergence with migration - first period.
    Ti: Duration of divergence in isolation - second period.
    Tj: Duration of divergence with migration - third period.
    nr: Fraction of non-recombining regions, i.e. with reduced Ne.
    bf : Background factor that defines to which extent Ne is reduced.
    P: Fraction of the genome evolving neutrally.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Common demographic steps
    sts = moments.LinearSystem_1D.steady_state_1D(n1+n2)
    phi = moments.Spectrum(sts)
    phi = moments.Manips.split_1D_to_2D(phi, n1, n2)

    # Define a function for specific demographic steps
    def calc_bin(l_bf, l_nr, l_i, l_P):
        #l_bf is either = 1 or bf
        #l_nr is either = 1-nr or nr
        #l_i is either = 1 or 0
        #l_P is either = P or 1-P
        phiBin = copy.copy(phi)
        phiBin.integrate([nu1*l_bf, nu2*l_bf], Ts, m=numpy.array([[0, m12*l_i],[m21*l_i, 0]]))
        phiBin.integrate([nu1*l_bf, nu2*l_bf], Ti, m=numpy.array([[0, 0],[0, 0]]))
        phiBin.integrate([nu1*l_bf, nu2*l_bf], Tj, m=numpy.array([[0, m12*l_i],[m21*l_i, 0]]))
        fsBin = phiBin
        return( l_nr * l_P * fsBin)

    # Run the calc_bin function with a pool of nbcores agents - each core will have a value of bf, nr, i and P
    with Pool(processes=4) as pool:
        results = pool.starmap(calc_bin, zip([bf,1,bf,1], [nr,1-nr,nr,1-nr], [1,1,0,0], [P,P,1-P,1-P]))
        pool.close()
        pool.join()

    # Sum all spectra
    sumO = 0
    for binfs in results:
        sumO = sumO + binfs
        
    fs = sumO

    return fs

