#! /bin/bash
# genotype_refinement_raw_CF
# last edit: 2021-06
# authors: Christelle Fraïsse

#-# usage: qsub genotype_refinement_raw_CF.sh
#-# prog required: GATK v3.4-0-g7e26428
#-# files required: NamesTrio1.ped and NamesTrio2.ped because redundant individuals [a pedigree file produced with generate_PED.sh]
#-# description: Raw genotype refinement.


## define global variables
toGATK=/share/nas-isem_e/genomique_integrative/Cione_WGS
PROOT="/share/nas-isem_e/genomique_integrative/Cione_WGS"
refgen=/share/nas-isem_e/genomique_integrative/Cione_WGS/RefCiona/GCA_009617815.1_Ciona_intestinalis_HT_Hoya_T-line_assembly_2019_genomic.fna

## your joint sample
infolder=/share/nas-isem_e/genomique_integrative/Cione_WGS/assembly_2019/ciona_bwa-mapping_jointsamples_raw

## number of threads to use
nthr=30


##-1-## CalculateGenotypePosteriors: calculates genotype posterior likelihoods given panel data
## The genotype priors applied then follow a Dirichlet-Multinomial distribution, where 2 alleles per sample are drawn independently.
## This assumption of independent draws is the assumption Hardy-Weinberg Equilibrium. 
#  --supporting		NA	Other callsets to use in generating genotype posteriors
#  --deNovoPrior 	1.0E-6	The de novo mutation prior
#  --globalPrior 	0.001	The global Dirichlet prior parameters for the allele frequency
#  --ignoreInputSamples		Use external information only; do not inform genotype priors by the discovered allele frequency 
#				in the callset whose posteriors are being calculated. Useful for callsets containing related individuals.
#  --skipFamilyPriors		Skip application of family-based priors
#  --skipPopulationPriors	Skip application of population-based priors

#FORMAT
#JL phred-scaled joint likelihood of the genotype combination before FamilyPriors
#JP phred-scaled joint likelihood of the genotype combination after  FamilyPriors
#PP phred-scaled posterior probability

java -Xmx4g -jar ${toGATK}/GenomeAnalysisTK.jar \
	-T CalculateGenotypePosteriors \
	-R ${refgen} \
	-V ${infolder}/joint_bwa_mem_mdup_IR_raw.vcf \
	-o ${infolder}/joint_bwa_mem_mdup_IR_raw_refine_tmp.vcf \
	--skipPopulationPriors \
	-ped ${PROOT}/NamesTrio1.ped \

java -Xmx4g -jar ${toGATK}/GenomeAnalysisTK.jar \
        -T CalculateGenotypePosteriors \
        -R ${refgen} \
        -V ${infolder}/joint_bwa_mem_mdup_IR_raw_refine_tmp.vcf \
        -o ${infolder}/joint_bwa_mem_mdup_IR_raw_refine.vcf \
        --skipPopulationPriors \
        -ped ${PROOT}/NamesTrio2.ped \


##-2-## VariantFiltration
## Produce a filtered VCF of high-confidence calls
#  --mask			Input ROD mask
#  --filterExpression		One or more expression used with INFO fields to filter
#  --filterName			Names to use for the list of filters
#  --genotypeFilterExpression	One or more expression used with FORMAT (sample/genotype-level) fields to filter
#  --genotypeFilterName		Names to use for the list of sample/genotype filters (must be a 1-to-1 mapping)
#				this name is put in the FILTER field for variants that get filtered
#  --maskExtension 0		How many bases beyond records from a provided 'mask' rod should variants be filtered
#  --maskName			The text to put in the FILTER field if a 'mask' rod is provided and overlaps with a variant call
#  --filterNotInMask		Filter records NOT in given input mask.
#  -selectType			Select only a certain type of variants from the input file. Can be specified multiple times. Valid types are 
#				INDEL (insertion/deletion), SNP (single nucleotide polymorphism), MIXED (combination of SNPs and indels at a single position)
#				not recommended - MNP (multi-nucleotide polymorphism, e.g. a dinucleotide substitution) 
#				not recommended - SYMBOLIC (generally, a very large allele or one that's fuzzy and not fully modeled)
#				NO_VARIATION

#INFO
#QD Variant confidence/Quality by depth
#FS Phred-scaled p-value using Fisher's exact test to detect strand bias
#MQ RMS mapping quality
#MQRankSum Z-score from Wilcoxon rank sum test of Alt vs. Ref read mapping qualities
#ReadPosRankSum Z-score from Wilcoxon rank sum test of Alt vs. Ref read position bias
#FORMAT
#FT genotype-level filter (PASS or lowGQ)
#FILTER
#variant-level filter (PASS or lowQ)  

# It's complaining that those values don't exist on particular lines. Which is entirely expected, since the RankSum metrics are only calculated on heterozygous sites.

# - apply filters to SNPs
java -Xmx4g -jar ${toGATK}/GenomeAnalysisTK.jar \
	-nt ${nthr} \
	-T SelectVariants \
	-R ${refgen} \
	--variant ${infolder}/joint_bwa_mem_mdup_IR_raw_refine.vcf \
	-o ${infolder}/joint_bwa_mem_mdup_IR_raw_refineSNP.vcf \
	-selectType SNP \

java -Xmx4g -jar ${toGATK}/GenomeAnalysisTK.jar \
	-T VariantFiltration \
	-R ${refgen} \
	--variant ${infolder}/joint_bwa_mem_mdup_IR_raw_refineSNP.vcf \
	-o ${infolder}/joint_bwa_mem_mdup_IR_raw_refineSNP_HQ.vcf \
	--filterExpression "QD < 10.0 || FS > 7.0 || MQ < 50.0 || MQRankSum < -1.5 || ReadPosRankSum < -1.5 || BaseQRankSum < -1.5 || BaseQRankSum > 1.5 || ClippingRankSum < -1.0 || ClippingRankSum > 1.0 " \
	--filterName "lowQ" \
	--genotypeFilterExpression "GQ < 35.0" \
	--genotypeFilterName "lowGQ" \

# - apply filters to Indels
java -Xmx4g -jar ${toGATK}/GenomeAnalysisTK.jar \
	-nt ${nthr} \
	-T SelectVariants \
	-R ${refgen} \
	--variant ${infolder}/joint_bwa_mem_mdup_IR_raw_refine.vcf \
	-o ${infolder}/joint_bwa_mem_mdup_IR_raw_refineINDEL.vcf \
	-selectType INDEL \

java -Xmx4g -jar ${toGATK}/GenomeAnalysisTK.jar \
	-T VariantFiltration \
	-R ${refgen} \
	--variant ${infolder}/joint_bwa_mem_mdup_IR_raw_refineINDEL.vcf \
	-o ${infolder}/joint_bwa_mem_mdup_IR_raw_refineINDEL_HQ.vcf \
	--filterExpression "QD < 10.0 || FS > 7.0 || ReadPosRankSum < -1.5" \
	--filterName "lowQ" \
	--genotypeFilterExpression "GQ < 35.0" \
	--genotypeFilterName "lowGQ" \


##-3-## CombineVariants: Combines VCF records from different sources.
#  -V					VCF files (ROD Tracks)
#  --filteredrecordsmergetype 	NA	Determines how we should handle records seen at the same site in the VCF, but with different FILTER fields
#					KEEP_IF_ANY_UNFILTERED Union - leaves the record if any record is unfiltered.
#					KEEP_IF_ALL_UNFILTERED Requires all records present at site to be unfiltered.
#					KEEP_UNCONDITIONAL If any record is present at this site (regardless of possibly being filtered),
#					then all such records are kept and the filters are reset.
#  --genotypemergeoption 	NA	Determines how we should merge genotype records for the same sample
#					UNIQUIFY Make all sample genotypes unique by file. Each sample shared across RODs gets named sample.ROD.
#					PRIORITIZE Take genotypes in priority order (see the priority argument).
#					UNSORTED Take the genotypes in any order.
#					REQUIRE_UNIQUE Require that all samples/genotypes be unique between all inputs. 
#  -priority				A comma-separated string describing the priority ordering for the genotypes as far as which record gets emitted
#  --minimumN 1				Combine variants and output site only if the variant is present in at least N input files
#  --assumeIdenticalSamples	NA	If true, assume input VCFs have identical sample sets and disjoint calls
#  --excludeNonVariants 	NA	Don't include loci found to be non-variant after the combining procedure		
#  --filteredAreUncalled 	NA	If true, then filtered VCFs are treated as uncalled, so that filtered set annotations don't appear in the combined VCF

java -Xmx4g -jar ${toGATK}/GenomeAnalysisTK.jar \
	-nt ${nthr} \
	-R ${refgen} \
	-T CombineVariants \
	-o ${infolder}/joint_bwa_mem_mdup_IR_raw_refine_HQ.vcf \
	-V:snp ${infolder}/joint_bwa_mem_mdup_IR_raw_refineSNP_HQ.vcf \
	-V:indel ${infolder}/joint_bwa_mem_mdup_IR_raw_refineINDEL_HQ.vcf \
	--filteredrecordsmergetype KEEP_IF_ALL_UNFILTERED \
	--genotypemergeoption PRIORITIZE \
	-priority snp,indel \


echo "# finished Raw Genotype Refinement with GATK"


##-4-## SelectVariants
#  -V					VCF files (ROD Tracks)
#  -select 		NA		One or more criteria to use when selecting the data
#					'AF > 0.01'
#					'DP > 10.0'
#					set == "Intersection": occurred in both call sets, not filtered out
#					set == "NAME": occurred in the call set NAME only
#					set=NAME1-filteredInNAME: occurred in both call sets, but was not filtered in NAME1 but was filtered in NAME2
#					set=filteredInAll: occurred in both call sets, but was filtered out of both
#					set=NAME1-NAME2: a variant occurred in both NAME1 and NAME2 but not all sets
#  --sample_file 	NA		File containing a list of samples (one per line) to include. Can be specified multiple times
#  --concordance 	NA		Output variants that were also called in this comparison track
#  --discordance 	NA		Output variants that were not called in this comparison track
#  --maxIndelSize 	2147483647	indel size select
#  --mendelianViolationQualThreshold 0.0	Minimum genotype QUAL score for each trio member required to accept a site as a violation
#  --restrictAllelesTo 	ALL		Select only variants of a particular allelicity (ALL, MULTIALLELIC or BIALLELIC)
#  --selectTypeToInclude NA		Select only a certain type of variants from the input file (INDEL, SNP, MIXED, MNP, SYMBOLIC, NO_VARIATION)
#  --excludeFiltered 	NA		Don't include filtered loci in the analysis
#  --excludeNonVariants	NA		Exclude sites monomorphic after sample subsetting
#  --mendelianViolation NA		output mendelian violation sites only

java -Xmx4g -jar ${toGATK}/GenomeAnalysisTK.jar \
	-nt ${nthr} \
	-R ${refgen} \
	-T SelectVariants \
	--variant ${infolder}/joint_bwa_mem_mdup_IR_raw_refine_HQ.vcf \
	-o ${infolder}/dbase.bwa_mem_mdup_IR_raw_refine_HQ.vcf \
	--excludeFiltered \


echo "# Raw Database creation with GATK"
