#! /bin/bash
# snpindel_recalibration_CF
# last edit: 2021-06
# authors: Christelle Fraïsse

#-# usage: qsub snpindel_recalibration_CF.sh
#-# prog required: GATK v3.4-0-g7e26428
# 		   VQSR recalibration plots: install.packages("gsalib"); install.packages("ggplot2"); install.packages("reshape")
#-# files required: dbase.bwa_mem_mdup_IR_raw_refine_HQ.vcf [high-quality variant database]
#-# description: SNP & INDEL recalibration.


## define global variables
toGATK=/share/nas-isem_e/genomique_integrative/Cione_WGS
PROOT="/share/nas-isem_e/genomique_integrative/Cione_WGS"
refgen=/share/nas-isem_e/genomique_integrative/Cione_WGS/RefCiona/GCA_009617815.1_Ciona_intestinalis_HT_Hoya_T-line_assembly_2019_genomic.fna

## your joint sample
infolder=/share/nas-isem_e/genomique_integrative/Cione_WGS/assembly_2019/ciona_bwa-mapping_jointsamples_raw

## your variant database
dbfolder=/share/nas-isem_e/genomique_integrative/Cione_WGS/assembly_2019/ciona_bwa-mapping_jointsamples_raw

## number of threads to use
nthr=30


##### Variant Quality Score Recalibration
##-1-## Build the SNP recalibration model
## VariantRecalibrator takes the overlap of the training/truth resource sets and of your callset. 
## It models the distribution of these variants relative to the annotations specified, and attempts to group them into clusters.
## Then it uses the clustering to assign VQSLOD scores to all variants in the INFO field.
#  -an			The names of the annotations which should used for calculations
#			If some annotations are missing use VariantAnnotator
#  -resource		A list of sites for which to apply a prior prob of being correct but which aren't used by the algorithm (training and truth sets required)
#  -recal_file		The output recal file used by ApplyRecalibration
#  -tranches_file	The output tranches file used by ApplyRecalibration
#  -rscript_file	The output rscript file generated by the VQSR to aid in visualization of the input data and learned model
#  -mode		Recalibration mode to employ
#  --ignore_all_filters	The variant recalibrator will ignore all input filters. Useful to rerun the VQSR from a filtered output file
#  --maxGaussians 8	Max number of Gaussians for the positive model

#INFO
#VQSLOD Log odds ratio of being a true variant versus being false under the trained gaussian mixture model

java -Xmx30g -jar ${toGATK}/GenomeAnalysisTK.jar \
	-nt ${nthr} \
	-T VariantRecalibrator \
	-R ${refgen} \
	-input ${infolder}/joint_bwa_mem_mdup_IR_raw.vcf \
	-resource:DB_HQ,known=false,training=true,truth=true,prior=10.0 ${dbfolder}/dbase.bwa_mem_mdup_IR_raw_refine_HQ.vcf \
	-an DP -an QD -an FS -an SOR -an MQ -an MQRankSum -an ReadPosRankSum \
	-mode SNP \
	-tranche 100.0 -tranche 99.9 -tranche 99.0 -tranche 90.0 \
	-recalFile ${infolder}/joint_bwa_mem_mdup_IR_recal_SNP.recal \
	-tranchesFile ${infolder}/joint_bwa_mem_mdup_IR_recal_SNP.tranches \
	-rscriptFile ${infolder}/joint_bwa_mem_mdup_IR_recal_SNP_plots.R \
	

##-2-## Apply SNP recalibration
## ApplyRecalibration applies the model to each variant producing a recalibrated VCF in which variants are annotated with VQSLOD.
## In addition, the program looks at what is the VQSLOD value above which 99.0% of the variants in the training callset are included. 
## It then takes that value of VQSLOD and uses it as a threshold to filter the variants (FILTER field will have the name of the tranche they belonged to).
## The first tranche -lowest values- is exceedingly specific, i.e. few false positive, but less sensitive, i.e. many false negative.
#  -recal_file		The output recal file used by ApplyRecalibration
#  -tranches_file	The output tranches file used by ApplyRecalibration
#  -mode		Recalibration mode to employ: 
#			1.) SNP for recalibrating only SNPs (emitting indels untouched in the output VCF)
#			2.) INDEL for indels
#			3.) BOTH for recalibrating both SNPs and indels simultaneously
#  --excludeFiltered	Don't output filtered loci after applying the recalibration
#  --ignore_all_filters	If specified, the variant recalibrator will ignore all input filters. Useful to rerun the VQSR from a filtered output file
#  --ts_filter_level	The truth sensitivity level at which to start filtering

java -Xmx30g -jar ${toGATK}/GenomeAnalysisTK.jar \
	-nt ${nthr} \
	-T ApplyRecalibration \
	-R ${refgen} \
	-input ${infolder}/joint_bwa_mem_mdup_IR_raw.vcf \
	-mode SNP \
	--ts_filter_level 99.5 \
	-recalFile ${infolder}/joint_bwa_mem_mdup_IR_recal_SNP.recal \
	-tranchesFile ${infolder}/joint_bwa_mem_mdup_IR_recal_SNP.tranches \
	-o ${infolder}/joint_bwa_mem_mdup_IR_recal_snps_raw_indels.vcf \


##-3-## Build the INDEL recalibration model
java -Xmx30g -jar ${toGATK}/GenomeAnalysisTK.jar \
	-nt ${nthr} \
	-T VariantRecalibrator \
	-R ${refgen} \
	--maxGaussians 4 \
	-input ${infolder}/joint_bwa_mem_mdup_IR_recal_snps_raw_indels.vcf \
	-resource:DB_HQ,known=false,training=true,truth=true,prior=12.0 ${dbfolder}/dbase.bwa_mem_mdup_IR_raw_refine_HQ.vcf \
	-an DP -an QD -an FS -an SOR -an MQRankSum -an ReadPosRankSum \
	-mode INDEL \
	-tranche 100.0 -tranche 99.9 -tranche 99.0 -tranche 90.0 \
	-recalFile ${infolder}/joint_bwa_mem_mdup_IR_recal_INDEL.recal \
	-tranchesFile ${infolder}/joint_bwa_mem_mdup_IR_recal_INDEL.tranches \
	-rscriptFile ${infolder}/joint_bwa_mem_mdup_IR_recal_INDEL_plots.R \


##-4-## Apply INDEL recalibration
java -Xmx30g -jar ${toGATK}/GenomeAnalysisTK.jar \
	-nt ${nthr} \
	-T ApplyRecalibration \
	-R ${refgen} \
	-input ${infolder}/joint_bwa_mem_mdup_IR_recal_snps_raw_indels.vcf \
	-mode INDEL \
	--ts_filter_level 99.0 \
	-recalFile ${infolder}/joint_bwa_mem_mdup_IR_recal_INDEL.recal \
	-tranchesFile ${infolder}/joint_bwa_mem_mdup_IR_recal_INDEL.tranches \
	-o ${infolder}/joint_bwa_mem_mdup_IR_recal_variants.vcf \


echo "# finished Variant Recalibration with GATK"
