#! /bin/bash
# genotype_refinement_recal_CF
# last edit: 2021-06
# authors: Christelle Fraïsse

#-# usage: qsub genotype_refinement_recal_CF.sh
#-# prog required: GATK v3.4-0-g7e26428
#-# files required: NamesTrio1.ped and NamesTrio2.ped because redundant individuals [a pedigree file produced with generate_PED.sh] 
#-# description: Genotype refinement after recalibration.


## define global variables
toGATK=/share/nas-isem_e/genomique_integrative/Cione_WGS
PROOT="/share/nas-isem_e/genomique_integrative/Cione_WGS"
refgen=/share/nas-isem_e/genomique_integrative/Cione_WGS/RefCiona/GCA_009617815.1_Ciona_intestinalis_HT_Hoya_T-line_assembly_2019_genomic.fna

## your joint sample
infolder=/share/nas-isem_e/genomique_integrative/Cione_WGS/assembly_2019/ciona_bwa-mapping_jointsamples_raw

## number of threads to use
nthr=30


##-1-## Derive posterior probabilities of genotypes
## CalculateGenotypePosteriors calculates genotype posterior likelihoods given panel data
#  --supporting		NA	Other callsets to use in generating genotype posteriors
#  --deNovoPrior 	1.0E-6	The de novo mutation prior
#  --globalPrior 	0.001	The global Dirichlet prior parameters for the allele frequency
#  --ignoreInputSamples		Use external information only; do not inform genotype priors by the discovered allele frequency 
#				in the callset whose posteriors are being calculated. Useful for callsets containing related individuals.
#  --skipFamilyPriors		Skip application of family-based priors
#  --skipPopulationPriors	Skip application of population-based priors

#FORMAT
#JL phred-scaled joint likelihood of the genotype combination before FamilyPriors
#JP phred-scaled joint likelihood of the genotype combination after  FamilyPriors
#PP phred-scaled posterior probability

java -Xmx30g -jar ${toGATK}/GenomeAnalysisTK.jar \
	-T CalculateGenotypePosteriors \
	-R ${refgen} \
	-V ${infolder}/joint_bwa_mem_mdup_IR_recal_variants.vcf \
	-o ${infolder}/joint_bwa_mem_mdup_IR_recal_variants_refine_tmp.vcf \
	--skipPopulationPriors \
	-ped ${PROOT}/NamesTrio1.ped \

java -Xmx30g -jar ${toGATK}/GenomeAnalysisTK.jar \
        -T CalculateGenotypePosteriors \
        -R ${refgen} \
        -V ${infolder}/joint_bwa_mem_mdup_IR_recal_variants_refine_tmp.vcf \
        -o ${infolder}/joint_bwa_mem_mdup_IR_recal_variants_refine.vcf \
        --skipPopulationPriors \
        -ped ${PROOT}/NamesTrio2.ped \


##-2-## Filter low quality genotypes
## VariantFiltration produces a filtered VCF of high-confidence calls
#  --mask			Input ROD mask
#  --filterExpression		One or more expression used with INFO fields to filter
#  --filterName			Names to use for the list of filters
#  --genotypeFilterExpression	One or more expression used with FORMAT (sample/genotype-level) fields to filter
#  --genotypeFilterName		Names to use for the list of sample/genotype filters (must be a 1-to-1 mapping)
#				this name is put in the FILTER field for variants that get filtered
# --maskExtension 0		How many bases beyond records from a provided 'mask' rod should variants be filtered
# --maskName			The text to put in the FILTER field if a 'mask' rod is provided and overlaps with a variant call
# --filterNotInMask		Filter records NOT in given input mask.

#INFO
#QD Variant confidence/Quality by depth
#FS Phred-scaled p-value using Fisher's exact test to detect strand bias
#MQ RMS mapping quality
#MQRankSum Z-score from Wilcoxon rank sum test of Alt vs. Ref read mapping qualities
#ReadPosRankSum Z-score from Wilcoxon rank sum test of Alt vs. Ref read position bias
#FORMAT
#FT genotype-level filter (PASS or lowGQ)

java -Xmx30g -jar ${toGATK}/GenomeAnalysisTK.jar \
	-T VariantFiltration \
	-R ${refgen} \
	--variant ${infolder}/joint_bwa_mem_mdup_IR_recal_variants_refine.vcf \
	-o ${infolder}/joint_bwa_mem_mdup_IR_recal_variants_refine_HQ.vcf \
	--filterExpression "QUAL < 30.0" --filterName "lowQ" \
	--genotypeFilterExpression "GQ < 30.0" --genotypeFilterName "lowGQ" \
	--genotypeFilterExpression "DP < 10" --genotypeFilterName "lowGQ" \


##-3-## Annotate possible de novo mutations
## VariantAnnotator
#  -A		One or more specific annotations to apply to variant calls
#  -E		One or more specific expressions to apply to variant calls
#  -G		One or more classes/groups of annotations to apply to variant calls
#  -mvq 0.0	The genotype quality threshold in order to annotate mendelian violation ratio

java -Xmx30g -jar ${toGATK}/GenomeAnalysisTK.jar \
	-nt ${nthr} \
	-T VariantAnnotator \
	-R ${refgen} \
	--variant ${infolder}/joint_bwa_mem_mdup_IR_recal_variants_refine_HQ.vcf \
	-o ${infolder}/joint_bwa_mem_mdup_IR_recal_variants_refine_HQ_denovo_tmp.vcf \
	-A PossibleDeNovo \
	-ped ${PROOT}/NamesTrio1.ped \

java -Xmx30g -jar ${toGATK}/GenomeAnalysisTK.jar \
        -nt ${nthr} \
        -T VariantAnnotator \
        -R ${refgen} \
        --variant ${infolder}/joint_bwa_mem_mdup_IR_recal_variants_refine_HQ_denovo_tmp.vcf \
        -o ${infolder}/joint_bwa_mem_mdup_IR_recal_variants_refine_HQ_denovo.vcf \
        -A PossibleDeNovo \
        -ped ${PROOT}/NamesTrio2.ped \


echo "# finished Genotype Refinement with GATK"
